#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
import time
from dataclasses import dataclass, field
from typing import cast, List
import pytest
from frocket.common.config import config
from frocket.common.dataset import DatasetInfo
from frocket.common.helpers.utils import timestamped_uuid
from frocket.common.tasks.async_tracker import AsyncJobTracker, AsyncJobStage
from frocket.common.tasks.base import BaseJobResult, TaskStatus, JobStats
from frocket.common.tasks.query import QueryJobResult
from frocket.common.tasks.registration import RegisterArgs, DatasetValidationMode, REGISTER_DEFAULT_VALIDATION_MODE, \
    REGISTER_DEFAULT_VALIDATE_UNIQUES, RegistrationJobResult
from frocket.datastore.registered_datastores import get_datastore
from frocket.invoker import invoker_api
from frocket.invoker.stats_builder import TASK_COMPLETION_GRANULARITY_SECONDS
from tests.utils.base_test_utils import SKIP_LAMBDA_TESTS
from tests.utils.dataset_utils import TestDatasetInfo, DEFAULT_GROUP_COLUMN, DEFAULT_TIMESTAMP_COLUMN, TestColumn, \
    BASE_TIME, DEFAULT_GROUP_COUNT, DEFAULT_ROW_COUNT, new_test_dataset
# noinspection PyUnresolvedReferences
from tests.utils.redis_fixture import init_test_redis_settings
# noinspection PyUnresolvedReferences
from tests.utils.lambda_fixture import init_mock_lambda_settings

# TODO skip this whole module if no docker env/processes running - using a class?
# TODO test query/register timeout (sync/async)
# TODO Test S3 permission denied - how to simulate with mock? (probably should be test_registration_job)

ORIGINAL_INVOKER_TYPE = config['invoker']
INVOKER_TYPES = ['work_queue']
if not SKIP_LAMBDA_TESTS:
    INVOKER_TYPES.append('aws_lambda')
API_DATASET_NUM_PARTS = 3


@pytest.fixture(scope="session", autouse=True)
def set_invoker_intervals():
    config['invoker.retry.failed.interval'] = '0.05'


@pytest.fixture(params=INVOKER_TYPES)
def curr_invoker_type(request):
    config['invoker'] = request.param
    print(f"Running test with invoker type set to: {request.param}")
    yield request.param
    config['invoker'] = ORIGINAL_INVOKER_TYPE


@dataclass
class ApiDatasetInfo:
    test_ds: TestDatasetInfo
    remote_basepath: str
    datasets_to_clean: List[str] = field(default_factory=list)
    default_dataset: DatasetInfo = field(default=None)

    def build_register_args(self,
                            name_prefix: str = None,
                            override_basepath: str = None,
                            group_id_column: str = DEFAULT_GROUP_COLUMN,
                            validation_mode: DatasetValidationMode = REGISTER_DEFAULT_VALIDATION_MODE,
                            validate_uniques: bool = REGISTER_DEFAULT_VALIDATE_UNIQUES) -> RegisterArgs:
        return RegisterArgs(name=timestamped_uuid(name_prefix),
                            basepath=override_basepath or self.remote_basepath,
                            group_id_column=group_id_column,
                            timestamp_column=DEFAULT_TIMESTAMP_COLUMN,
                            validation_mode=validation_mode,
                            validate_uniques=validate_uniques)

    def _async_run(self, tracker: AsyncJobTracker) -> BaseJobResult:
        statuses = list(tracker.generator())
        for st in statuses:
            print(st)

        assert statuses[0].stage in [AsyncJobStage.STARTING, AsyncJobStage.RUNNING]
        assert statuses[-2].stage in [AsyncJobStage.FINISHING, AsyncJobStage.RUNNING]
        assert statuses[-1].stage == AsyncJobStage.DONE

        # Ensure that all "mid" stages are RUNNING, and that the no. of ENDED_SUCCESS can only increase
        first_running_idx = 0 if statuses[0].stage == AsyncJobStage.RUNNING else 1
        last_running_idx = -2 if statuses[-2].stage == AsyncJobStage.RUNNING else -3

        run_stages = statuses[first_running_idx:(last_running_idx + 1)]
        run_stages_count = len(run_stages)
        assert run_stages_count >= 1
        for i, status in enumerate(run_stages):
            assert status.stage == AsyncJobStage.RUNNING
            prev_status = run_stages[i - 1] if i > 0 else None
            if status.task_counters:
                num_tasks = sum(status.task_counters.values())
                assert 1 <= num_tasks <= API_DATASET_NUM_PARTS  # Exact no. of parts not exposed by invoker_api
                if prev_status and prev_status.task_counters:
                    assert status.task_counters.get(TaskStatus.ENDED_SUCCESS, 0) >= \
                           prev_status.task_counters.get(TaskStatus.ENDED_SUCCESS, 0)
        return statuses[-1].result

    def register(self, args: RegisterArgs = None,
                 run_async: bool = False,
                 expected_success_value: bool = True,
                 expected_basepath: str = None,
                 expected_size: int = None) -> RegistrationJobResult:
        args = args or self.build_register_args()
        self.datasets_to_clean.append(args.name)
        start_time = time.time()
        if run_async:
            tracker = invoker_api.register_dataset_async(args, set_max_wait=True)
            result = self._async_run(tracker)
        else:
            result = invoker_api.register_dataset(args)
        total_time = time.time() - start_time
        assert type(result) == RegistrationJobResult
        result = cast(RegistrationJobResult, result)
        if expected_success_value:
            assert result.success
            assert result.error_message is None
            assert result.dataset.total_parts == API_DATASET_NUM_PARTS

            expected_size = expected_size or self.test_ds.expected_parts.total_size
            expected_total_tasks = 1 if args.validation_mode == DatasetValidationMode.SINGLE else 2
            self.check_stats(result.stats, expected_total_tasks, expected_size, total_time)

            if expected_basepath:
                assert result.dataset.basepath == expected_basepath
        else:
            assert not result.success
            assert not result.dataset
        return result

    def get_default_dataset(self) -> DatasetInfo:
        if not self.default_dataset:
            self.default_dataset = self.register().dataset
        return self.default_dataset

    def query(self,
              query: dict,
              override_dataset: DatasetInfo = None,
              expected_size: int = None,
              run_async: bool = False,
              expected_success_value: bool = True) -> QueryJobResult:
        ds = override_dataset or self.get_default_dataset()
        start_time = time.time()
        if run_async:
            tracker = invoker_api.run_query_async(dataset=ds, query=query, set_max_wait=True)
            result = self._async_run(tracker)
        else:
            result = invoker_api.run_query(ds, query)
        total_time = time.time() - start_time
        assert type(result) == QueryJobResult
        result = cast(QueryJobResult, result)

        if expected_success_value:
            assert result.success
            assert result.error_message is None
            assert result.query
            expected_parts = override_dataset.total_parts if override_dataset else self.default_dataset.total_parts
            expected_size = expected_size or self.test_ds.expected_parts.total_size
            self.check_stats(result.stats,
                             expected_total_tasks=expected_parts,
                             expected_size=expected_size,
                             total_time=total_time)
        else:
            assert not result.success
            assert not result.query
            assert not result.funnel
        return result

    def check_stats(self, stats: JobStats, expected_total_tasks: int, expected_size: int, total_time: float):
        # Part count & size
        assert stats.dataset.parts == API_DATASET_NUM_PARTS
        assert stats.dataset.total_size == expected_size
        # Task count
        assert stats.invoker.total_tasks == expected_total_tasks and stats.invoker.failed_tasks == 0
        assert sum(stats.invoker.task_success_over_time.values()) == expected_total_tasks
        assert stats.worker.cold_tasks + stats.worker.warm_tasks == expected_total_tasks
        assert sum(stats.worker.cache.values()) == expected_total_tasks
        # Timing (0.5 / 0.25 of total were set low numbers for sanity, but CI tests are still flaky)
        assert total_time >= stats.total_time > 0  # TODO backlog fix flakiness in > (total_time * 0.5)
        assert total_time >= (stats.invoker.enqueue_time +
                              stats.invoker.poll_time) > 0  # TODO backlog fix flakiness in > (total_time * 0.25)
        max_key_in_completion_dict = \
            math.ceil(total_time / TASK_COMPLETION_GRANULARITY_SECONDS) * TASK_COMPLETION_GRANULARITY_SECONDS
        assert all([0 < ts <= max_key_in_completion_dict
                    for ts in stats.invoker.task_success_over_time.keys()])
        # Just a most basic sanity for the TimingStats
        for timing in [stats.worker.invoke_latency, stats.worker.load_time, stats.worker.total_time]:
            assert timing is not None
            assert all(0 < t < total_time for t in timing.values())
            assert min(timing.values()) == timing['min'] and max(timing.values()) == timing['max']
            assert timing['max'] >= timing['50%'] >= timing['min']

    def cleanup(self):
        self.test_ds.cleanup()
        for ds_name in self.datasets_to_clean:
            invoker_api.unregister_dataset(ds_name, force=True)

    @classmethod
    def instance(cls):
        if not hasattr(cls, '_shared_instance'):
            print("Creating ApiDatasetInfo shared instance (should happen once)")
            test_ds = TestDatasetInfo.build(parts=API_DATASET_NUM_PARTS)
            s3path = test_ds.copy_to_s3()
            info = ApiDatasetInfo(test_ds=test_ds,
                                  remote_basepath=s3path)
            cls._shared_instance = info
        return cls._shared_instance


@pytest.fixture(scope='module')
def api_dataset():
    info = cast(ApiDatasetInfo, ApiDatasetInfo.instance())
    yield info
    info.cleanup()


def test_register_remote(api_dataset: ApiDatasetInfo, curr_invoker_type):
    common_args = api_dataset.build_register_args()
    # Register local dataset
    result = api_dataset.register(common_args, expected_basepath=api_dataset.remote_basepath)
    first_id = result.dataset.id

    # Now re-register with same args (== same name), check that it works and registered_at is newer
    result = api_dataset.register(common_args, expected_basepath=api_dataset.remote_basepath)
    second_id = result.dataset.id

    assert first_id != second_id and \
           first_id.name == second_id.name and \
           first_id.registered_at < second_id.registered_at


def test_register_invalid_path(api_dataset: ApiDatasetInfo):
    args = api_dataset.build_register_args(override_basepath='/thisisnotadirectory')
    result = api_dataset.register(args, expected_success_value=False)

    args = api_dataset.build_register_args(override_basepath='/root')
    result = api_dataset.register(args, expected_success_value=False)

    args = api_dataset.build_register_args(override_basepath='s3://thisisnotabucket/')
    result = api_dataset.register(args, expected_success_value=False)


def test_register_invalid_column(api_dataset: ApiDatasetInfo, curr_invoker_type):
    args = api_dataset.build_register_args(group_id_column='notacolumn',
                                           validation_mode=DatasetValidationMode.SINGLE, validate_uniques=False)
    result = api_dataset.register(args, expected_success_value=False)


def test_register_async(api_dataset: ApiDatasetInfo, curr_invoker_type):
    result = api_dataset.register(run_async=True)


def test_register_async_fail(api_dataset: ApiDatasetInfo, curr_invoker_type):
    args = api_dataset.build_register_args(group_id_column='notacolumn',
                                           validation_mode=DatasetValidationMode.SINGLE, validate_uniques=False)
    result = api_dataset.register(args, run_async=True, expected_success_value=False)


def test_get_dataset_missing_none():
    result = invoker_api.get_dataset(name="notreallyman", throw_if_missing=False)
    assert result is None


@pytest.mark.xfail(strict=True)
def test_get_dataset_missing_xfail():
    _ = invoker_api.get_dataset(name="notreallyman", throw_if_missing=True)


def test_get_dataset(api_dataset: ApiDatasetInfo):
    job_result = api_dataset.register(api_dataset.build_register_args(name_prefix="getme-"))
    assert job_result.dataset.id.name.startswith("getme-")
    ds = invoker_api.get_dataset(job_result.dataset.id.name, throw_if_missing=True)
    assert ds == job_result.dataset


def test_get_dataset_detailed(api_dataset: ApiDatasetInfo):
    job_result = api_dataset.register()
    ds = invoker_api.get_dataset(job_result.dataset.id.name)
    short_schema = invoker_api.get_dataset_schema(ds, full=False)
    full_schema = invoker_api.get_dataset_schema(ds, full=True)
    parts_info = invoker_api.get_dataset_parts(ds)
    # The details of these objects are tested elsewhere, here we're mainly testing that invoker_api fetches them...
    assert short_schema == get_datastore().short_schema(ds)
    assert full_schema == get_datastore().schema(ds)
    assert parts_info == api_dataset.test_ds.expected_parts and parts_info == get_datastore().dataset_parts_info(ds)


def test_unregister_no_usage(api_dataset: ApiDatasetInfo):
    def unreg(force: bool):
        ds_name = api_dataset.register().dataset.id.name
        unreg_result = invoker_api.unregister_dataset(ds_name, force=force)
        assert unreg_result.success and unreg_result.error_message is None
        assert unreg_result.dataset_found and unreg_result.dataset_last_used is None

        ds = invoker_api.get_dataset(ds_name, throw_if_missing=False)
        assert not ds

        unreg_result = invoker_api.unregister_dataset(ds_name, force=force)
        assert unreg_result.success and unreg_result.error_message is None
        assert not unreg_result.dataset_found and unreg_result.dataset_last_used is None

    unreg(force=False)
    unreg(force=True)


def test_unregister_after_use_force(api_dataset: ApiDatasetInfo):
    # fail, force, wait, set to zero...
    ds = api_dataset.register().dataset
    api_dataset.query(query={}, override_dataset=ds, )

    unreg_result = invoker_api.unregister_dataset(ds.id.name)
    assert not unreg_result.success
    assert invoker_api.get_dataset(ds.id.name, throw_if_missing=True)

    unreg_result = invoker_api.unregister_dataset(ds.id.name, force=True)
    assert unreg_result.success and unreg_result.dataset_found and unreg_result.dataset_last_used
    assert invoker_api.get_dataset(ds.id.name, throw_if_missing=False) is None


def test_unregister_after_use_wait(api_dataset: ApiDatasetInfo):
    orig_interval = config['unregister.last.used.interval']
    try:
        config['unregister.last.used.interval'] = '100'
        ds = api_dataset.register().dataset
        start = time.time()
        api_dataset.query(query={}, override_dataset=ds)
        unreg_result = invoker_api.unregister_dataset(ds.id.name)
        assert not unreg_result.success

        time_passed = time.time() - start
        next_second = int(math.ceil(time_passed))
        config['unregister.last.used.interval'] = str(next_second)
        time.sleep(next_second - time_passed + 0.2)
        unreg_result = invoker_api.unregister_dataset(ds.id.name)
        assert unreg_result.success
    finally:
        config['unregister.last.used.interval'] = orig_interval


def test_list_datasets(api_dataset: ApiDatasetInfo):
    original_list = invoker_api.list_datasets()

    ds1_args = api_dataset.build_register_args()
    ds1 = api_dataset.register(ds1_args).dataset
    ds2 = api_dataset.register().dataset
    updated_list = invoker_api.list_datasets()
    assert len(updated_list) == len(original_list) + 2
    for e in [*original_list, ds1, ds2]:
        assert e in updated_list

    # Re-register dataset with same name
    ds1new = api_dataset.register(ds1_args).dataset
    updated_list = invoker_api.list_datasets()
    assert len(updated_list) == len(original_list) + 2
    for e in [*original_list, ds1new, ds2]:
        assert e in updated_list

    assert invoker_api.unregister_dataset(ds1new.id.name).success
    assert invoker_api.unregister_dataset(ds2.id.name).success
    updated_list = invoker_api.list_datasets()
    assert len(updated_list) == len(original_list)
    assert sorted(updated_list, key=lambda d: d.id.name) == sorted(original_list, key=lambda d: d.id.name)


def run_and_validate_query(api_dataset: ApiDatasetInfo, run_async: bool):
    query = {
        'query': {
            'conditions': [
                {'filter': [TestColumn.int_64_ts.value, '>=', BASE_TIME]}
            ]
        },
        'funnel': {
            'sequence': [
                {'filter': [TestColumn.int_64_ts.value, '==', BASE_TIME]}
            ],
            'endAggregations': [
                {'column': TestColumn.str_userid, 'type': 'count'}
            ]
        }
    }

    query_result = api_dataset.query(query, run_async=run_async)
    print(query_result.to_json(indent=2))
    assert query_result.query.matching_groups == DEFAULT_GROUP_COUNT * API_DATASET_NUM_PARTS
    assert query_result.query.matching_group_rows == DEFAULT_ROW_COUNT * API_DATASET_NUM_PARTS
    assert query_result.query.aggregations is None
    assert query_result.stats.worker.scanned_groups == DEFAULT_GROUP_COUNT * API_DATASET_NUM_PARTS
    assert query_result.stats.worker.scanned_rows == DEFAULT_ROW_COUNT * API_DATASET_NUM_PARTS

    assert len(query_result.funnel.sequence) == 1
    assert query_result.funnel.sequence[0].matching_groups == 1
    funnel_row_count = query_result.funnel.sequence[0].matching_group_rows
    assert 1 <= funnel_row_count < DEFAULT_ROW_COUNT
    assert query_result.funnel.sequence[0].aggregations is None

    assert len(query_result.funnel.end_aggregations) == 1
    assert query_result.funnel.end_aggregations[0].value == funnel_row_count


def test_query(api_dataset: ApiDatasetInfo, curr_invoker_type):
    run_and_validate_query(api_dataset, run_async=False)


def test_query_async(api_dataset: ApiDatasetInfo, curr_invoker_type):
    iterations = 5  # A bit more certainty on our asserts around async job progress
    for _ in range(iterations):
        run_and_validate_query(api_dataset, run_async=True)


@pytest.mark.xfail(strict=True)
def test_query_invalid_xfail(api_dataset: ApiDatasetInfo):
    query = {'query': {'conditions': [{'filter': [TestColumn.int_64_ts.value, 'contains', BASE_TIME]}]}}
    api_dataset.query(query)


def run_query_missing_files(api_dataset: ApiDatasetInfo, run_async: bool):
    with new_test_dataset(API_DATASET_NUM_PARTS) as test_ds:
        s3path = test_ds.copy_to_s3()
        query = {}

        register_args = RegisterArgs(
            name=timestamped_uuid(),
            basepath=s3path,
            group_id_column=DEFAULT_GROUP_COLUMN,
            timestamp_column=DEFAULT_TIMESTAMP_COLUMN)

        ds = api_dataset.register(register_args,
                                  expected_basepath=s3path,
                                  expected_size=test_ds.expected_parts.total_size).dataset
        api_dataset.query(query,
                          override_dataset=ds,
                          expected_size=test_ds.expected_parts.total_size,
                          expected_success_value=True)

        # Re-register the dataset, so workers won't use disk cache but rather fetch the (missing) files from source
        ds = api_dataset.register(register_args,
                                  expected_basepath=s3path,
                                  expected_size=test_ds.expected_parts.total_size).dataset
        for key in test_ds.bucket.objects.all():
            key.delete()
        api_dataset.query(query, override_dataset=ds, expected_success_value=False)


def test_query_missing_files_failure(api_dataset: ApiDatasetInfo, curr_invoker_type):
    run_query_missing_files(api_dataset, run_async=False)


def test_query_async_fail(api_dataset: ApiDatasetInfo, curr_invoker_type):
    run_query_missing_files(api_dataset, run_async=True)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import math
import time
from enum import Enum, auto
from typing import Callable, Any, Dict, cast, Union, Type, Optional
import redis
from redis import Redis
from frocket.common.config import config
from frocket.common.serializable import Envelope, SerializableDataClass
from frocket.common.tasks.base import TaskStatus, BaseTaskResult, TaskAttemptId, TaskStatusUpdate, \
    BaseTaskRequest, BlobId
from frocket.common.dataset import DatasetInfo, DatasetPartsInfo, DatasetPartId, DatasetShortSchema, DatasetSchema
from frocket.common.helpers.utils import timestamped_uuid
from frocket.datastore.datastore import Datastore, WorkerSelectedPart, DEFAULT_QUEUE, DEFAULT_DEQUEUE_WAIT_TIME
from frocket.datastore.blobstore import Blobstore, BLOB_DEFAULT_TTL, BLOB_MAX_TTL

logger = logging.getLogger(__name__)

FROCKET_KEY_PREFIX = config.get('datastore.redis.prefix')
DATASETS_HASH_KEY = f"{FROCKET_KEY_PREFIX}datasets"
DS_PARTS_HASH_KEY = f"{FROCKET_KEY_PREFIX}dataset_parts"
DS_LAST_USED_HASH_KEY = f"{FROCKET_KEY_PREFIX}dataset_lastused"
SCHEMAS_HASH_KEY = f"{FROCKET_KEY_PREFIX}schemas"
SHORT_SCHEMAS_HASH_KEY = f"{FROCKET_KEY_PREFIX}short_schemas"
QUEUES_PREFIX = f"{FROCKET_KEY_PREFIX}queues"
BLOB_KEY_PREFIX = f"{FROCKET_KEY_PREFIX}blob"


class RedisStore(Datastore, Blobstore):
    """
    Implementation of both Datastore and Blobstore interfaces with Redis.

    Note that each of these interfaces can be configured to use its own host/port/logical DB, or they could use the
    same one (see configuration guide).

    The underlying Redis object is thread-safe, and since this class has no mutable state,
    it is also considered thread-safe. However, running in a multi-threaded model has not been seriously tested yet.
    """
    ConverterFunc = Callable[[str], Any]

    class ScopedKey(Enum):
        TASK_ATTEMPT_STATUS = auto()
        TASK_ATTEMPT_RESULT = auto()
        TASK_ATTEMPT_NO = auto()
        JOB_PUBLISHED_PARTS = auto()

        # noinspection PyUnusedLocal
        def __init__(self, *args):
            self._name_for_key = self.name.lower().replace('_', ':')

        def build(self, reqid: str, suffix: Any = None) -> str:
            res = f"{FROCKET_KEY_PREFIX}req-{reqid}:{self._name_for_key}"
            if suffix:
                res += f":{suffix}"
            return res

    _redis: redis.Redis
    _host: str

    def __init__(self, role=Union["blobstore", "datastore"]):
        self._role = role
        self._host = config.get_with_fallbacks(f"{role}.redis.host", "redis.host")
        self._port = int(config.get_with_fallbacks(f"{role}.redis.port", "redis.port"))
        self._db = int(config.get_with_fallbacks(f"{role}.redis.db", "redis.db"))

        self._redis = redis.Redis(host=self._host, port=self._port, db=self._db)
        self._redis.ping()  # Just make sure it works

    def write_dataset_info(self, dataset, parts, schema):
        # When re-registering a dataset (with a newer timestamp), it overrides the previous one -
        # That's why the key in datasets hash is by name only.
        # Related objects are written in the same transaction to prevent corruption
        with self._redis.pipeline() as pipe:
            dskey = dataset.id.name
            pipe.hset(DATASETS_HASH_KEY, dskey, dataset.to_json())
            pipe.hset(DS_PARTS_HASH_KEY, dskey, parts.to_json())
            pipe.hset(SCHEMAS_HASH_KEY, dskey, schema.to_json())  # TODO backlog make more compact for storage
            pipe.hset(SHORT_SCHEMAS_HASH_KEY, dskey, schema.short().to_json())
            pipe.hdel(DS_LAST_USED_HASH_KEY, dskey)
            pipe.execute()

    def remove_dataset_info(self, name):
        with self._redis.pipeline() as pipe:
            dskey = name
            deleted_count = pipe.hdel(DATASETS_HASH_KEY, dskey)
            pipe.hdel(DS_PARTS_HASH_KEY, dskey)
            pipe.hdel(SCHEMAS_HASH_KEY, dskey)
            pipe.hdel(SHORT_SCHEMAS_HASH_KEY, dskey)
            pipe.hdel(DS_LAST_USED_HASH_KEY, dskey)
            pipe.execute()
        return deleted_count != 0

    def dataset_info(self, name):
        return self._read_dataset_data(name, DATASETS_HASH_KEY, DatasetInfo)

    def dataset_parts_info(self, ds):
        return self._read_dataset_data(ds.id.name, DS_PARTS_HASH_KEY, DatasetPartsInfo)

    def schema(self, ds):
        return self._read_dataset_data(ds.id.name, SCHEMAS_HASH_KEY, DatasetSchema)

    def short_schema(self, ds):
        return self._read_dataset_data(ds.id.name, SHORT_SCHEMAS_HASH_KEY, DatasetShortSchema)

    def _read_dataset_data(self, dataset_name: str, key: str,
                           cls: Type[SerializableDataClass]) -> Optional[SerializableDataClass]:
        res = self._redis.hget(key, dataset_name)
        return cls.from_json(res) if res else None

    # TODO backlog monitor and limit queue size (if no workers are active to consume it, or they fail to start)
    def enqueue(self, requests, queue=DEFAULT_QUEUE):
        queue_key = f"{QUEUES_PREFIX}:{queue}"

        with self._redis.pipeline() as pipe:
            for req in requests:
                pipe.lpush(queue_key, Envelope.seal_to_json(req))  # See Envelope class - enables polymorphism
            pipe.execute()

    def dequeue(self, queue=DEFAULT_QUEUE, timeout=DEFAULT_DEQUEUE_WAIT_TIME):
        queue_key = f"{QUEUES_PREFIX}:{queue}"

        json_string = None
        if timeout > 0:
            res = self._redis.brpop(queue_key, timeout)
            if res:
                json_string = res[1].decode()
        else:
            res = self._redis.rpop(queue_key)
            if res:
                json_string = res.decode()

        if json_string:
            obj = Envelope.open_from_json(json_string, expected_superclass=BaseTaskRequest)
            return obj
        else:
            return None

    def update_task_status(self, reqid, tasks, status):
        if type(tasks) is TaskAttemptId:
            self._write_task_status(reqid, tasks, status)
        elif type(tasks) is list:
            with self._redis.pipeline() as pipe:
                for taskid in tasks:
                    self._write_task_status(reqid, taskid, status, conn=pipe)
                pipe.execute()
        else:
            raise Exception(f"Unknown type: {type(tasks)}")

    def _write_task_status(self, reqid: str, taskid: TaskAttemptId, status: TaskStatus, conn: Redis = None):
        conn = conn or self._redis
        conn.hset(self.ScopedKey.TASK_ATTEMPT_STATUS.build(reqid),
                  taskid.to_json(),
                  TaskStatusUpdate.now(status).to_json())

    def tasks_status(self, reqid):
        return self._read_hash_for_request(self.ScopedKey.TASK_ATTEMPT_STATUS, reqid,
                                           lambda k: TaskAttemptId.from_json(k),
                                           lambda v: TaskStatusUpdate.from_json(v))

    def write_task_result(self, reqid, taskid, result):
        with self._redis.pipeline() as pipe:
            pipe.hset(self.ScopedKey.TASK_ATTEMPT_RESULT.build(reqid),
                      taskid.to_json(),
                      Envelope.seal_to_json(result))
            self._write_task_status(reqid, taskid, result.status, conn=pipe)
            pipe.execute()

    def task_results(self, reqid):
        return self._read_hash_for_request(self.ScopedKey.TASK_ATTEMPT_RESULT, reqid,
                                           lambda k: TaskAttemptId.from_json(k),
                                           lambda v: Envelope.open_from_json(v, expected_superclass=BaseTaskResult))

    def datasets(self):
        items = self._read_hash(DATASETS_HASH_KEY,
                                lambda k: k,  # Un-needed
                                lambda v: DatasetInfo.from_json(v))
        return list(items.values())

    def last_used(self, ds: DatasetInfo) -> int:
        res = self._redis.hget(DS_LAST_USED_HASH_KEY, ds.id.name)
        return int(res) if res else None

    def mark_used(self, ds: DatasetInfo):
        self._redis.hset(DS_LAST_USED_HASH_KEY, ds.id.name, math.floor(time.time()))
        pass

    def _read_hash_for_request(self, scoped_key: ScopedKey, reqid: str,
                               key_reader: ConverterFunc, value_reader: ConverterFunc) -> Dict:
        return self._read_hash(scoped_key.build(reqid), key_reader, value_reader)

    def _read_hash(self, key: str, key_reader: ConverterFunc, value_reader: ConverterFunc) -> Dict:
        items_dict = self._redis.hgetall(key)
        if items_dict:
            return {key_reader(k.decode('utf-8')): value_reader(v.decode('utf-8'))
                    for k, v in items_dict.items()}
        else:
            return {}

    def publish_for_worker_selection(self, reqid, attempt_round, parts):
        set_key = self.ScopedKey.JOB_PUBLISHED_PARTS.build(reqid, suffix=attempt_round)
        string_members = [part.to_json() for part in parts]
        self._redis.sadd(set_key, *string_members)

    def self_select_part(self, reqid, attempt_round, candidates=None):
        set_key = self.ScopedKey.JOB_PUBLISHED_PARTS.build(reqid, attempt_round)
        # Try selecting one of this worker's candidates, if any
        if candidates:
            for candidate in candidates:
                cand_string = candidate.to_json()
                removed_count = self._redis.srem(set_key, cand_string)
                if removed_count == 1:
                    return WorkerSelectedPart(part_id=candidate, random=False, task_attempt_no=attempt_round)

        res = self._redis.spop(set_key)  # Fallback to taking a random task
        if res:
            part_id = cast(DatasetPartId, DatasetPartId.from_json(res.decode()))
            return WorkerSelectedPart(part_id=part_id, random=True, task_attempt_no=attempt_round)
        else:
            return None

    def increment_attempt(self, reqid, partidx):
        new_attempt_no = self._redis.hincrby(self.ScopedKey.TASK_ATTEMPT_NO.build(reqid), str(partidx))
        if not new_attempt_no:
            raise Exception(f"Failed incrementing task attempt no. for request {reqid} part {partidx}")

        return int(new_attempt_no)

    def cleanup_request_data(self, reqid: str) -> None:
        # Cleanup all scoped keys. Not all necessarily exist.
        # (also with suffixes for JOB_PUBLISHED_PARTS: see method self.publish_for_worker_selection)
        keys = [sk.build(reqid) for sk in self.ScopedKey] + \
               [self.ScopedKey.JOB_PUBLISHED_PARTS.build(reqid, suffix=i) for i in range(3)]
        self._redis.delete(*keys)

    #
    # Blobstore methods
    #

    def write_blob(self, data, ttl=None, tag=None):
        ttl = ttl or BLOB_DEFAULT_TTL
        if BLOB_MAX_TTL != 0 and (ttl == 0 or ttl > BLOB_MAX_TTL):
            raise Exception(f"Requested TTL {ttl} is longer than maximum: {BLOB_MAX_TTL}")

        prefix = f"{BLOB_KEY_PREFIX}:{tag or ''}-"
        new_id = timestamped_uuid(prefix)
        self._redis.setex(new_id, ttl, data)
        return new_id

    @staticmethod
    def _validate_blob_id(blobid: BlobId):
        if not blobid.startswith(BLOB_KEY_PREFIX):
            raise Exception(f"Invalid blob ID: {blobid}")

    def read_blob(self, blobid):
        self._validate_blob_id(blobid)
        return self._redis.get(blobid)

    def delete_blob(self, blobid):
        self._validate_blob_id(blobid)
        deleted_count = self._redis.delete(blobid)
        return deleted_count == 1

    def __str__(self):
        return f"RedisStore(role {self._role}, host {self._host}:{self._port}, db {self._db})"



"""
While metrics support in Funnel Rocket is built with Prometheus (or more generally OpenMetrics) in mind,
all Prometheus-specific code is in this module.

TODO backlog support help string (documentation) per each member in MetricName enum
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Dict, Type
from prometheus_client import Counter, Histogram
from prometheus_client.metrics import MetricWrapperBase
from frocket.common.config import config
from frocket.common.helpers.utils import memoize
from frocket.common.metrics import MetricName, MeasuredUnit, supported_label_names, MetricData, empty_label_names

prom_counters: Dict[MetricName, Counter] = {}
prom_histograms: Dict[MetricName, Histogram] = {}


@memoize
def buckets_by_unit(unit: MeasuredUnit) -> List[float]:
    """Each unit (seconds, bytes, dollars) may have its own buckets configured, or fallback to the default."""
    assert unit is not MeasuredUnit.COUNT  # COUNT should not use a histogram
    buckets_string = config.get_with_fallbacks(f'metrics.buckets.{unit.name.lower()}', 'metrics.buckets.default')
    buckets = [float(b) for b in buckets_string.split(',')]
    return buckets


def unit_to_metric_type(unit: MeasuredUnit) -> Type[MetricWrapperBase]:
    """The type of Prometheus metric is automatically derived from the type of measured unit."""
    if unit is MeasuredUnit.COUNT:
        return Counter
    else:
        return Histogram


def init_prom_metrics():
    """In Prometheus clients, all metrics should be defined only once before use, along with their possible labels.
    This is not a technical limitation of Prometheus itself, but rather enforced by official clients."""
    for e in MetricName:
        base_args = {'name': e.name.lower(),
                     'documentation': e.name,
                     'labelnames': supported_label_names(e)}
        metric_type = unit_to_metric_type(e)
        if metric_type == Counter:
            prom_counters[e] = Counter(**base_args)
        elif metric_type == Histogram:
            prom_histograms[e] = Histogram(**base_args, buckets=buckets_by_unit(e.unit))


def update(metrics: List[MetricData]):
    """Update (increment/observe) new values after a job completes, etc."""
    for md in metrics:
        empty_labels = empty_label_names(md.name)
        all_labels = {**empty_labels, **md.labels}
        metric_type = unit_to_metric_type(md.name.unit)
        if metric_type == Counter:
            prom_counters[md.name].labels(**all_labels).inc(md.value)
        elif metric_type == Histogram:
            prom_histograms[md.name].labels(**all_labels).observe(md.value)



import pathlib
import setuptools

this_dir = pathlib.Path(__file__).parent
requirements_file = this_dir / "requirements.txt"
readme_file = this_dir / "README.md"

install_requires = requirements_file.read_text().splitlines()
long_description = readme_file.read_text() if readme_file.exists() else ''

setuptools.setup(
    name="funnel-rocket",
    version="0.5.3",
    author="Elad Rosenheim, Avshalom Manevich",
    author_email="elad@dynamicyield.com",
    description="Cloud native distributed funnel queries",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DynamicYieldProjects/funnel-rocket-oss",
    packages=setuptools.find_packages(),
    package_data={
        "frocket": ["resources/*.*"],
    },
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=install_requires
)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from enum import auto
from frocket.common.serializable import AutoNamedEnum


class ValidationErrorKind(AutoNamedEnum):
    """Distinguish between types of validation issues in query"""
    INVALID_ARGUMENTS = auto()  # Validator given wrong arguments
    SCHEMA = auto()  # Failure at JSON Schema level
    TYPE_MISMATCH = auto()  # Operator or value type don't match each other, or the context
    DATASET_MISMATCH = auto()  # Column names, types, etc. do not match the schema of the given dataset
    RELATION = auto()  # query.relation expression found invalid by relation_parser.py
    # Note for unexpected errors: unlike other kinds, the message associated with this kind may leak sensitive data
    # if it was returned to the caller - so it is not returned by the API server in PUBLIC mode.
    UNEXPECTED = auto()


class QueryValidationError(Exception):
    def __init__(self, message: str, kind: ValidationErrorKind = None):
        self.message = message
        self.kind = kind or ValidationErrorKind.UNEXPECTED  # Default, but should be rare.

    @staticmethod
    def wrap(e: Exception, kind: ValidationErrorKind = None):
        return QueryValidationError(str(e), kind)

    def __str__(self):
        return f"ValidationError({self.kind.value}: {self.message})"



"""
Transform a given list of metrics from multiple sources (invoker, workers) into one DataFrame, for easy analysis.
Export the data into file and/or Prometheus, by configuration.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import List, Dict, Union
import pandas as pd
from pandas import DataFrame
from frocket.common.config import config
from frocket.common.metrics import SourceAndMetricTuple, ALL_LABEL_NAMES
from frocket.invoker import prom_adapter

logger = logging.getLogger(__name__)

METRIC_SOURCE_COLUMN = 'source'
METRIC_NAME_COLUMN = 'metric'
METRIC_VALUE_COLUMN = 'value'

PANDAS_FLOAT_FORMAT = '{:.5f}'  # No pesky scientific notation ;-)
pd.options.display.float_format = PANDAS_FLOAT_FORMAT.format

# 'Last run' file, if defines, stores the most recent job's metrics as a file in CSV or Parquet format (by extension)
EXPORT_LASTRUN_FILE = config.get('metrics.export.lastrun', None)
EXPORT_TO_PROMETHEUS = config.bool('metrics.export.prometheus')

if EXPORT_TO_PROMETHEUS:
    prom_adapter.init_prom_metrics()


class MetricsFrame:
    def __init__(self, source_and_metrics: List[SourceAndMetricTuple]):
        self._sources = [ms.source for ms in source_and_metrics]
        self._metrics = [ms.metric for ms in source_and_metrics]
        self._build_df()

    def _build_df(self):
        """
        Build the DataFrame: each row is one reported metric, but the DF is created with columns. Hence, we're creating
        columns here rather than rows.
        """
        metric_source_column = self._sources
        metric_name_column = [m.name.name for m in self._metrics]  # Metric names column
        metric_value_column = [m.value for m in self._metrics]     # Metric values column

        # Init empty columns for all possible label names.
        # Cells not not filled (see below) will remain empty (and possibly even entire columns)
        label_columns: Dict[str, List[Union[str, None]]] = {}
        for label_name in ALL_LABEL_NAMES:
            label_columns[label_name] = [None] * len(self._metrics)

        # Fill labels columns with what labels are actually set per metric
        for i, metric in enumerate(self._metrics):
            for label_name, label_value in metric.labels.items():
                label_columns[label_name][i] = label_value

        df_columns = {METRIC_SOURCE_COLUMN: metric_source_column,
                      METRIC_NAME_COLUMN: metric_name_column,
                      METRIC_VALUE_COLUMN: metric_value_column,
                      **label_columns}
        self._df = pd.DataFrame(data=df_columns)
        # logger.debug(f"Types: {self._df.dtypes.index.tolist()}, data:\n{self._df}")  # If needed

    def export(self) -> None:
        if EXPORT_LASTRUN_FILE:
            self._to_lastrun_file(EXPORT_LASTRUN_FILE)
        if EXPORT_TO_PROMETHEUS:
            self._to_prometheus()

    def _to_prometheus(self) -> None:
        prom_adapter.update(self._metrics)

    def _to_lastrun_file(self, filename: str) -> None:
        if filename.lower().endswith('.parquet'):
            self._df.to_parquet(filename, index=False)
        else:
            self._df.to_csv(filename, float_format=PANDAS_FLOAT_FORMAT, index=False)

    @property
    def dataframe(self) -> DataFrame:
        return self._df



"""
Invoke tasks by enqueing them in the datastore. Not much to do here :-)
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from frocket.invoker.impl.async_invoker import AsyncInvoker


class WorkQueueInvoker(AsyncInvoker):
    def _enqueue(self, requests) -> None:
        self._datastore.enqueue(requests)



#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from abc import abstractmethod, ABCMeta
from typing import Dict, Counter, List, NamedTuple, Optional
from frocket.common.metrics import MetricsBag, ComponentLabel, MetricName, MetricData, SourceAndMetricTuple, LabelsDict
from frocket.common.tasks.base import TaskAttemptId, TaskStatus, BaseTaskResult, \
    BaseTaskRequest, BaseJobResult, JobStatus, JobStats
from frocket.common.tasks.async_tracker import AsyncJobStatusUpdater, AsyncJobStage
from frocket.common.helpers.utils import timestamped_uuid
from frocket.datastore.registered_datastores import get_datastore
from frocket.invoker.jobs.job import Job
from frocket.invoker.metrics_frame import MetricsFrame
from frocket.invoker.stats_builder import build_stats
from frocket.worker.impl.generic_env_metrics import GenericEnvMetricsProvider

logger = logging.getLogger(__name__)


class BaseInvoker(metaclass=ABCMeta):
    """
    BaseInvoker defined the basic logic of invoking jobs using the concrete Job subclass it's initialized with.

    However, this class does not contain the logic for how to actually 'get workers to do stuff' and collect results,
    which can possibly be done in various ways. That part is handled by AsyncInvoker and its own subclasses
    (WorkQueueInvoker, AwsLambdaInvoker, and future ones). It is ofc possible to also implement a SyncInvoker subclass
    (and as its subclasses have either regular Lambda invocation, calls to HTTP APIs, etc.)

    See also the Job class documentation for the workflow from the job's perspective.
    """
    def __init__(self, job: Job):
        self._job = job
        # Generate a new unique request ID and also hand it to the job instance
        self._request_id = self.new_request_id()
        self._job.request_id = self._request_id
        self._datastore = get_datastore()
        self._metrics = MetricsBag(component=ComponentLabel.INVOKER,
                                   env_metrics_provider=GenericEnvMetricsProvider())

    # noinspection PyBroadException
    def run(self, async_status_updater: AsyncJobStatusUpdater = None) -> BaseJobResult:
        try:
            # The run itself
            with self._metrics.measure(MetricName.INVOKER_TOTAL_SECONDS):
                final_job_status, all_task_results, latest_task_results = \
                    self._run_to_completion(async_status_updater)

            # Post-run: collect metrics, build job result
            job_labels = self._job.metric_labels
            final_metrics = self._metrics.finalize(success=final_job_status.success,
                                                   added_labels=job_labels)
            # Note that metrics are collected for all task attempts, successful and failed
            metrics_frame = self.__build_metrics(job_labels=job_labels,
                                                 invoker_metrics=final_metrics,
                                                 all_task_results=all_task_results)
            run_result = self.__build_run_result(final_job_status=final_job_status,
                                                 latest_task_results=latest_task_results,
                                                 metrics_frame=metrics_frame)
            metrics_frame.export()
        except Exception:
            try:
                final_job_status = JobStatus(success=False, error_message='Unexpected error', attempts_status=[])
                run_result = self.__build_run_result(final_job_status=final_job_status,
                                                     latest_task_results=[],
                                                     metrics_frame=None)
                logger.exception('Run failed')
            except Exception:
                logger.exception('Failed to set a "success=False" result via the job builder')
                run_result = BaseJobResult(success=False, error_message='Unexpected error',
                                           request_id=self._request_id, stats=JobStats())

        # In case some lost tasks are still running, they may still "resurrect" some keys with their own data -
        # though the job is done as far as the invoker is concerned.
        # TODO backlog periodic "old junk" cleaner (can look for outdated request IDs, which are timestamped)
        self._datastore.cleanup_request_data(self._request_id)

        if async_status_updater:
            async_status_updater.done(run_result)
        return run_result

    class RunCompletionResult(NamedTuple):
        final_job_status: JobStatus
        all_task_results: Dict[TaskAttemptId, BaseTaskResult] = {}
        latest_task_results: List[BaseTaskResult] = []

    def _run_to_completion(self, async_status_updater) -> RunCompletionResult:
        # First step: call the job to run any preparation code which needs to run invoker-side,
        # might take a bit of time and might also fail (e.g. dataset files discovery)
        try:
            error_message = self._job.prerun(async_status_updater)
        except Exception as e:
            logger.exception(f"Prerun failed: {e}")
            error_message = str(e)

        if error_message:  # Failed already at pre-run, before any tasks were invoked
            final_job_status = JobStatus(success=False, error_message=error_message, attempts_status=[])
            return self.RunCompletionResult(final_job_status)

        # Second step: build tasks, run them till done successfully or retries exhausted, and gather results

        task_requests = self.__build_initial_tasks()
        if async_status_updater:
            async_status_updater.update(stage=AsyncJobStage.RUNNING)
        job_status = self._do_run(task_requests, async_status_updater)  # Calls subclass!
        all_task_results = self._datastore.task_results(self._request_id)  # Read results

        # Third step: finishing up: sanity check, and call job again now that results are in.
        # The job can now process and/or validate task results taken together, and may still fail the run.

        if async_status_updater:
            async_status_updater.update(stage=AsyncJobStage.FINISHING)

        latest_task_results = self.latest_results_only(job_status, all_task_results)
        if job_status.success:
            successful_tasks = [tr for tr in latest_task_results if tr.status == TaskStatus.ENDED_SUCCESS]
            assert (len(successful_tasks) == self._job.total_tasks())

        final_job_status = self._job.complete(job_status, latest_task_results, async_status_updater)
        return self.RunCompletionResult(final_job_status=final_job_status,
                                        all_task_results=all_task_results,
                                        latest_task_results=latest_task_results)

    @abstractmethod
    def _do_run(self,
                task_requests: List[BaseTaskRequest],
                async_status_updater: AsyncJobStatusUpdater = None) -> JobStatus:
        """Actually invoke the tasks and return with a status when they are complete."""
        pass

    def __build_initial_tasks(self) -> List[BaseTaskRequest]:
        requests = self._job.build_tasks()
        # If relevant, publish dataset parts for self-selection by workers
        parts_to_publish = self._job.dataset_parts_to_publish()
        if parts_to_publish:
            assert all([req.invoker_set_task_index is None for req in requests])

        # All tasks' status is always set to QUEUED before actually invoking them
        attempt_ids = [TaskAttemptId(task_index=req.invoker_set_task_index or i) for i, req in enumerate(requests)]
        self._datastore.update_task_status(self._request_id, attempt_ids, TaskStatus.QUEUED)

        if parts_to_publish:
            self._datastore.publish_for_worker_selection(self._request_id,
                                                         parts=parts_to_publish,
                                                         attempt_round=0)

        return requests

    def _build_retry_task(self, task_index) -> (TaskAttemptId, BaseTaskRequest):
        attempt_no = self._datastore.increment_attempt(self._request_id, task_index)
        attempt_id = TaskAttemptId(task_index, attempt_no)

        req = self._job.build_retry_task(attempt_no, task_index)
        self._datastore.update_task_status(self._request_id, attempt_id, TaskStatus.QUEUED)
        return req

    def __build_run_result(self,
                           final_job_status: JobStatus,
                           latest_task_results: List[BaseTaskResult],
                           metrics_frame: Optional[MetricsFrame]) -> BaseJobResult:

        # Collect errors from invoker and tasks (only from latest attempts)
        full_error_message = final_job_status.error_message
        task_errors: List[str] = [res.error_message for res in latest_task_results if res.error_message]
        if len(task_errors) > 0:
            full_error_message = f"{full_error_message or 'No errors in invoker'}. Task errors: {task_errors}"
        stats = build_stats(metrics_frame, self._job.parts_info()) if metrics_frame else None

        # Collect the common 'base' attributes and pass them to the job, to use as **args in init'ing the result object
        base_attributes = BaseJobResult(
            request_id=self._request_id,
            success=final_job_status.success,
            error_message=full_error_message,
            stats=stats).\
            shallowdict(include_none=True)

        run_result = self._job.build_result(
            base_attributes, final_job_status, latest_task_results)
        return run_result

    @staticmethod
    def __build_metrics(job_labels: LabelsDict,
                        invoker_metrics: List[MetricData],
                        all_task_results: Dict[TaskAttemptId, BaseTaskResult]) -> MetricsFrame:
        """Build a unified list of all reported metrics from the invoker and all tasks."""
        all_metrics: List[SourceAndMetricTuple] = \
            [SourceAndMetricTuple(source='invoker', metric=m) for m in invoker_metrics]

        for attempt_id, task_result in all_task_results.items():
            source_name = f"task-{attempt_id.task_index}-{attempt_id.attempt_no}"
            all_metrics += [SourceAndMetricTuple(source=source_name, metric=m.with_added_labels(job_labels))
                            for m in task_result.metrics]

        return MetricsFrame(all_metrics)

    @classmethod
    def latest_results_only(cls, job_status: JobStatus,
                            all_task_results: Dict[TaskAttemptId, BaseTaskResult]) -> List[BaseTaskResult]:
        """Get tasks results only for the single latest attempt per each task."""
        latest_attempts = [ts.latest_attempt for ts in job_status.attempts_status]
        results = [result for attempt_id, result in all_task_results.items()
                   if attempt_id in latest_attempts]
        # Prevent possible erratic behavior down the line by making order predictable (by task index)
        return sorted(results, key=lambda result: result.task_index)

    @classmethod
    def new_request_id(cls) -> str:
        return timestamped_uuid()

    @classmethod
    def simple_status_string(cls, status_counts: Counter[TaskStatus]) -> str:
        return ', '.join([f"{status.name}: {count}"
                          for status, count in status_counts.items()])



"""For everything but the kitchen sink."""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
import math
import random
import uuid
import time
from io import BytesIO
from typing import Optional, List
import pandas as pd
import pyarrow.feather as feather
import numpy as np


def terminal_red(message: str) -> str:
    return f"\033[31m{message}\033[0m"


def terminal_green(message: str) -> str:
    return f"\033[32m{message}\033[0m"


def memoize(obj):
    """Standard issue memoization decorator for caching function results (which don't need invalidation)."""
    cache = obj._cache = {}

    @functools.wraps(obj)
    def memoizer(*args, **kwargs):
        key = str(args) + str(kwargs)
        if key not in cache:
            cache[key] = obj(*args, **kwargs)
        return cache[key]

    return memoizer


def sample_from_range(range_max: int,
                      sample_ratio: float,
                      max_samples: int,
                      preselected: Optional[List[int]]) -> List[int]:
    """
    Given a range of numbers in 0..range_max, return random samples.
    Count of samples is set by sample_ratio, up to max_samples.
    If preselected is passed, include these indexes first.
    """
    available_indexes = list(range(range_max))
    sample_count = min(math.floor(range_max * sample_ratio), max_samples)

    if preselected:
        chosen = list(preselected)
        for i in preselected:
            available_indexes.remove(i)
        sample_count = max(sample_count - len(preselected), 0)
    else:
        chosen = []

    if sample_count > 0:
        chosen += random.choices(available_indexes, k=sample_count)
    return chosen


def timestamped_uuid(prefix: str = None) -> str:
    return f"{prefix or ''}{math.floor(time.time())}-{str(uuid.uuid4())[:8]}"


def ndarray_to_bytes(arr: np.ndarray) -> bytes:
    """Use PyArrow's feather format as a compute- and space-efficient format for serializing NumPy arrays."""
    df = pd.DataFrame(data={'arr': arr})
    buf = BytesIO()
    # noinspection PyTypeChecker
    feather.write_feather(df, buf)
    buf.seek(0)
    return buf.read()


def bytes_to_ndarray(data: bytes) -> np.ndarray:
    df = feather.read_feather(BytesIO(data))
    return df['arr']



"""
A collection of callback functions which the QueryValidator uses to extarct, validate and transform query elements,
with the kind help of PathVisitor class.

Functions which return a value are used to replace the given object with a different one,
which is handled by PathVisitor in its 'modifiable' mode.

Since callbacks are regular functions (not methods), and there's a bunch of them, they're in a separate file from
the QueryValidator class.

asserts are used where processing elements which should be already validated (so failures should be bugs).
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional
from frocket.common.validation.consts import DEFAULT_TARGET, AGGR_TYPES_WITH_OTHER_COLUMN, \
    DEFAULT_AGGREGATIONS, TARGET_TYPES_WITH_INCLUDE_ZERO, TARGET_OPS_SUPPORTING_INCLUDE_ZERO
from frocket.common.validation.error import ValidationErrorKind, QueryValidationError


def _to_verbose_filter(fltr) -> Optional[dict]:
    """If a condition filter is in short-hand notation (list), convert to verbose notation."""
    assert type(fltr) in [list, dict]
    if type(fltr) is list:
        assert len(fltr) == 3
        return {'column': fltr[0], 'op': fltr[1], 'value': fltr[2]}


def _to_verbose_target(target) -> Optional[dict]:
    """If a condition target is in short-hand notation (list), convert to verbose notation."""
    assert type(target) in [list, dict]
    if type(target) is list:
        assert len(target) in [3, 4]
        if len(target) == 3:
            return {'type': target[0], 'op': target[1], 'value': target[2]}
        elif len(target) == 4:
            return {'type': target[0], 'column': target[1], 'op': target[2], 'value': target[3]}


def _add_default_target(cond: dict) -> None:
    assert type(cond) is dict
    # (Modification is done on a key under the given object, so no need to return a modified dict)
    if ('filter' in cond or 'filters' in cond) and 'target' not in cond:  # Don't touch sequence conditions
        cond['target'] = DEFAULT_TARGET


def _validate_aggregation(aggr: dict) -> None:
    assert type(aggr) is dict
    aggr_type = aggr.get('type', None)
    other_column_required = aggr_type in AGGR_TYPES_WITH_OTHER_COLUMN
    other_column_found = 'otherColumn' in aggr

    if other_column_required != other_column_found:
        message = f"For aggregation {aggr} with type '{aggr_type}', other column name is "
        if other_column_required:
            message += 'required but was not found'
        else:
            message += 'not relevant but was given'
        raise QueryValidationError(message, kind=ValidationErrorKind.SCHEMA)


def _expand_aggregations(col_aggregations: list) -> Optional[list]:
    assert type(col_aggregations) is list
    result = []
    for aggr in col_aggregations:
        if aggr.get('type', None):
            result.append(aggr)
        else:
            if 'name' in aggr:
                message = f"Aggregation {aggr} expands into multiple default aggregations, " \
                          f"and thus a name attributeis not supported"
                raise QueryValidationError(message, kind=ValidationErrorKind.SCHEMA)
            for added_type in DEFAULT_AGGREGATIONS:
                result.append({**aggr, 'type': added_type})

    return result


def _validate_or_set_include_zero(cond: dict) -> None:
    """
    'includeZero' attribute of conditions may be tricky to get right.
    This function validates that its usage makes sense, and sets the correct default where it's ommitted.
    """
    assert type(cond) is dict
    if not ('filter' in cond or 'filters' in cond):
        return  # Skip sequence condition (and possibly other future types without a target)

    # This should run after _to_verbose_target() and _add_default_target() have already ran, ensuring target exists
    target_type = cond['target']['type']
    target_op = cond['target']['op']
    target_value = cond['target']['value']
    include_zero_value = cond.get('includeZero', None)
    target_as_string = f"{target_type} {target_op} {target_value}"

    if target_type not in TARGET_TYPES_WITH_INCLUDE_ZERO:
        if include_zero_value:  # Exists and set to True
            raise QueryValidationError(
                message=f"'includeZero' is not applicable for target type '{target_type}'. In condition: {cond}",
                kind=ValidationErrorKind.TYPE_MISMATCH)
    else:
        assert type(target_value) is int
        assert target_value >= 0

        if include_zero_value:  # Exists and set to True
            # Operator never relevant for includeZero=True
            if target_op not in TARGET_OPS_SUPPORTING_INCLUDE_ZERO:
                raise QueryValidationError(
                    message=f"For target operator '{target_op}', 'includeZero' cannot be true. In condition: {cond}",
                    kind=ValidationErrorKind.TYPE_MISMATCH)

            # Additional check when an operator is *potentially* relevant for includeZero=True
            if target_op == '<' and target_value == 0:
                raise QueryValidationError(
                    message=f"Target implies a negative value. In condition: {cond}",
                    kind=ValidationErrorKind.TYPE_MISMATCH)

            if (target_op == '!=' and target_value == 0) or \
                    (target_op in ['==', '>='] and target_value != 0):
                message = f"Target {target_as_string} explicitly precludes zero, and thus 'includeZero' " \
                          f"cannot be true. In condition: {cond}"
                raise QueryValidationError(message, kind=ValidationErrorKind.TYPE_MISMATCH)
        else:
            if target_op == '==' and target_value == 0:
                if include_zero_value is None:
                    # Explicitly set includeZero when target is count == 0
                    # Note: modifying a key under the given object, so no need to return a modified dict
                    cond['includeZero'] = True
                elif not include_zero_value:
                    message = f"When using a target of {target_as_string}, 'includeZero' cannot be false. " \
                                    f"Condition: {cond}"
                    raise QueryValidationError(message, kind=ValidationErrorKind.TYPE_MISMATCH)



"""
Simple CLI for Funnel Rocket.

This is currently a wrapper over invoker_api directly (meaning that the CLI process is the invoker), rather than
calling an API server - meaning that it does not rely on a running server, but needs the same permissions (listing files
in remote storage, access to Redis as datastore, optionally being able to invoke Lambdas).

This makes the CLI more suitable for onboarding and evaluation, but in production it's preferable to use the API
(for a better permissions model and centralized monitoring/logging, if nothing else).

The CLI does provide a few optional flags which make it also suitable for automating jobs:
* --nopretty returns JSON object/s without any captions
* --notrim and --nocolor prevents data from bein shortened or surrounded by ANSI color codes
* The log level is controllbable, and all log lines have a prefix making them easy to ignore.
"""
#  Copyright 2021 The Funnel Rocket Maintainers
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import argparse
# TODO backlog don't import any frocket modules but a carefully selected set which does not then import heavy packages
#  or initialize mechanisms. This is only partially done now (see import at end of file).
from frocket.common.config import config
from frocket.common.tasks.registration import DatasetValidationMode, REGISTER_DEFAULT_VALIDATION_MODE, \
    REGISTER_DEFAULT_FILENAME_PATTERN, REGISTER_DEFAULT_VALIDATE_UNIQUES

REGISTER_VALIDATION_MODE_CHOICES = [e.value.lower() for e in DatasetValidationMode]
LOG_LEVEL_CHOICES = ['debug', 'info', 'warning', 'error', 'critical']
LOG_LINE_PREFIX = '[Log '
LOG_FORMAT = LOG_LINE_PREFIX + '%(levelname)s %(name)s] %(message)s'


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description='Simple CLI for Funnel Rocket',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--notrim', action='store_true', help='Don\'t trim any text')
    parser.add_argument('--nocolor', action='store_true', help='Don\'t trim colorize any text')
    parser.add_argument('--nopretty', action='store_true', help='Don\'t pretty-print the response')
    parser.add_argument('--loglevel', type=str.lower, choices=LOG_LEVEL_CHOICES,
                        help=f'Set log level {LOG_LEVEL_CHOICES}')
    subparsers = parser.add_subparsers(dest='command', title='commands')
    subparsers.required = True

    register_parser = subparsers.add_parser('register', help='Register a dataset',
                                            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    register_parser.add_argument('name', type=str, help='Dataset name')
    register_parser.add_argument('basepath', type=str,
                                 help='The path all files are directly under. Local and s3://... paths supported.')
    register_parser.add_argument('group_id_column', type=str,
                                 help='The column to group rows by, e.g. "userId", "userHash". '
                                      'This column is required and no values can be missing. Each part (file) in the '
                                      'dataset should have a distinct set of values for this column.')
    register_parser.add_argument(
        'timestamp_column', type=str,
        help='The column holding the timestamp of each row, e.g. "timestamp", "ts". '
             'Must be a numeric column with no values missing. Using a unix timestamp is advised - '
             'with or without sub-second resoluton based on your needs, either as int or float.')
    register_parser.add_argument('--pattern', type=str, default=REGISTER_DEFAULT_FILENAME_PATTERN,
                                 help='Filename pattern. Sub-directories are currently not supported.')
    register_parser.add_argument('--validation', type=str.lower,
                                 choices=REGISTER_VALIDATION_MODE_CHOICES,
                                 default=REGISTER_DEFAULT_VALIDATION_MODE.value.lower(),
                                 help=f"Validation mode to use {REGISTER_VALIDATION_MODE_CHOICES}",
                                 metavar='MODE')
    register_parser.add_argument('--skip-uniques', action='store_true',
                                 default=not REGISTER_DEFAULT_VALIDATE_UNIQUES,
                                 help='Skip validation of group_id_column values uniqueness across files '
                                      '(the set of files to test is determined by --validation argument)')

    list_parser = subparsers.add_parser('list', help='List datasets')

    run_query_parser = subparsers.add_parser('run', help='Run query')
    run_query_parser.add_argument('dataset')
    query_sources_group = run_query_parser.add_mutually_exclusive_group(required=True)
    query_sources_group.add_argument('--file', '-f', type=str, help='Run query stored in file', dest='filename')
    query_sources_group.add_argument('--empty', '-e', action='store_true',
                                     help='Run an empty query with no conditions')
    query_sources_group.add_argument('--string' '-s', type=str,
                                     help='Run the following query string', dest='query_string')

    info_parser = subparsers.add_parser('info', help='Show dataset information')
    info_parser.add_argument('dataset', type=str)
    info_parser.add_argument('--full', action='store_true', help='Show full schema')

    unreg_parser = subparsers.add_parser('unregister', help='Unregister a dataset')
    unreg_parser.add_argument('dataset', type=str)
    unreg_parser.add_argument('--force', action='store_true',
                              help='Unregister a dataset even if it\'s currently in use')

    config_parser = subparsers.add_parser('config', help='Show configuration')
    return parser


def run_from_args(args: argparse.Namespace):
    config['log.format'] = LOG_FORMAT if args.nocolor else f"\033[33m{LOG_FORMAT}\033[0m"
    if args.loglevel:
        config['log.level'] = args.loglevel
    config.init_logging(force_console_output=True)

    # invoker_api isn't loaded (or logging implicitly initialized) till arguments are validated and log level is set
    from frocket.cli_commands import run_command
    run_command(args.command, args)


if __name__ == '__main__':
    parser = build_parser()
    args = parser.parse_args()
    run_from_args(args)
